# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import time
import subprocess

sys.path.insert(0, os.path.abspath('..'))

import adcc  # noqa: E402

# -- Project information -----------------------------------------------------


def determine_tag():
    commit = subprocess.check_output("git rev-parse HEAD".split(),
                                     universal_newlines=True).strip()
    tag = subprocess.check_output("git tag --points-at".split() + [f"{commit}"],
                                  universal_newlines=True).strip()
    return tag


project = 'adcc'
author = ", ".join(adcc.__authors__)
copyright = time.strftime("%Y") + ", " + author
version = adcc.__version__
release = adcc.__version__

if determine_tag() == "":
    version = "dev"
    release = adcc.__version__ + "+dev"

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.imgmath',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    # 'sphinx.ext.doctest',
    # 'sphinx.ext.todo',
    # 'sphinx.ext.coverage',
    #
    'sphinx_automodapi.automodapi',
    'sphinxcontrib.bibtex',
    'breathe'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst']

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# A nice logo. Should not exceed 200px
html_logo = None

# The canonical HTML url
html_baseurl = "https://adc-connect.org/"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "canonical_url": html_baseurl,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'adccdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'adcc.tex', 'adcc Documentation', author, 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'adcc', 'adcc Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'adcc', 'adcc Documentation',
     author, 'adcc', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# Imgmath settings
imgmath_latex_preamble = "".join([
    r"\newcommand*{\mat}[1]{\ensuremath{\mathbf{#1}}}",
    r"\renewcommand*{\vec}[1]{\ensuremath{\underline{\boldsymbol{#1}}}}"
    r"\DeclareMathOperator{\tr}{tr}",
    #
    r"\newcommand*{\Op}[1]{\ensuremath{\hat{\mathcal{#1}}}}",
    r"\newcommand*{\bra}[1]{\left\langle#1\right|}",
    r"\newcommand*{\ket}[1]{\left|#1\right\rangle}"
    r"\newcommand*{\braket}[2]{\left\langle#1\middle|#2\right\rangle}"
])

# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True

# __init__ docstring appears in documentation
autoclass_content = "init"

# Breathe settings
libadcc_doc_dir = (
    os.path.abspath(os.path.dirname(__file__))
    + "/../build/libadcc_docs/xml"
)
breathe_projects = {"libadcc": libadcc_doc_dir, }
breathe_default_project = "libadcc"

# Bibtex
bibtex_bibfiles = ['ref.bib', 'pub.bib']
